<?php

namespace Drupal\search_api_tax_filter\Plugin\search_api\processor;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformStateInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\search_api\Plugin\PluginFormTrait;
use Drupal\search_api\Processor\ProcessorPluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Excludes content associated with taxonomy terms from being indexed.
 *
 * @SearchApiProcessor(
 *   id = "search_api_tax_filter_processor",
 *   label = @Translation("Taxonomy Filter"),
 *   description = @Translation("Filter documents that are linked to the selected taxonomy terms to be indexed by search api."),
 *   stages = {
 *     "alter_items" = -50
 *   }
 * )
 */
class TaxonomyFilterProcessor extends ProcessorPluginBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The entity type bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected EntityTypeBundleInfoInterface $entityTypeBundleInfo;

  /**
   * Constructs a new SearchApiTaxFilterProcessor object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfoInterface $entity_type_bundle_info) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->entityTypeManager = $entity_type_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'vocabularies' => [],
      'terms' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $complete_form_state = $form_state instanceof SubformStateInterface ? $form_state->getCompleteFormState() : $form_state;

    $options = [];
    foreach ($this->entityTypeManager->getStorage('taxonomy_vocabulary')->loadMultiple() as $vocabulary) {
      $options[$vocabulary->id()] = $vocabulary->label();
    }
    $form['vocabularies'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Select the vocabularies to enable.'),
      '#options' => $options,
      '#default_value' => $this->configuration['vocabularies'] ?? [],
      '#ajax' => [
        'callback' => [$this, 'updateTerms'],
        'wrapper' => 'search-api-tax-filter-terms-wrapper',
      ],
    ];

    $target_bundles = $this->configuration['vocabularies'] ?? [];
    $parents = [
      'processors',
      'search_api_tax_filter_processor',
      'settings',
      'vocabularies',
    ];
    $triggering_element = $form_state->getTriggeringElement();
    if (isset($triggering_element['#parents']) && array_slice($triggering_element['#parents'], 0, -1) == $parents) {
      $values = $complete_form_state->getValues();
      $target_bundles = array_filter(NestedArray::getValue($values, $parents));
    }

    $form['terms'] = [
      '#type' => 'entity_autocomplete',
      '#title' => $this->t('Select Terms'),
      '#description' => $this->t('The terms chosen will be indexed, comma separated.'),
      '#target_type' => 'taxonomy_term',
      '#selection_settings' => [
        'target_bundles' => $target_bundles,
      ],
      '#default_value' => $this->entityTypeManager->getStorage('taxonomy_term')->loadMultiple($this->configuration['terms'] ?? []),
      '#tags' => TRUE,
      '#prefix' => '<div id="search-api-tax-filter-terms-wrapper">',
      '#suffix' => '</div>',
    ];

    return $form;
  }

  /**
   * AJAX callback to update the terms' selection.
   */
  public function updateTerms(array $form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    if (isset($triggering_element['#parents'])) {
      $parents = array_reverse(array_slice($triggering_element['#parents'], 1, -2));
      $parents[] = 'terms';
      return NestedArray::getValue($form, $parents);
    }

    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $this->setConfiguration([
      'vocabularies' => array_filter($values['vocabularies']),
      'terms' => array_map(function ($term) {
        return $term['target_id'];
      }, $values['terms']),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function alterIndexedItems(array &$items) {
    $config = $this->getConfiguration();

    $vocabularies = $config['vocabularies'];
    if (empty($vocabularies)) {
      return;
    }
    $terms = $config['terms'];
    if (empty($terms)) {
      return;
    }

    // Preprocess the terms to avoid complexity.
    $terms = $this->entityTypeManager->getStorage('taxonomy_term')->loadMultiple($terms);
    $allowed_terms = [];
    foreach ($terms as $term) {
      $allowed_terms[$term->bundle()][$term->id()] = $term->id();
    }

    $entity_type_ids = array_map(fn($datasource) => $datasource->getEntityTypeId(), $this->getIndex()->getDatasources());

    // Get the fields associated to the vocabularies.
    $fields = [];
    foreach ($entity_type_ids as $entity_type_id) {
      $bundle_info = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
      foreach (array_keys($bundle_info) as $bundle) {
        $fields = array_merge($fields, $this->getFieldNamesFromVocabularies($entity_type_id, $bundle, $vocabularies));
      }
    }
    $fields = array_unique($fields);

    // Avoid indexing any item that does not have the terms selected in the
    // right fields.
    foreach ($items as $item_id => $item) {
      $entity = $item->getOriginalObject()->getValue();
      foreach ($terms as $term) {
        $vocabulary_fields = $fields[$term->bundle()];
        $vocabulary_terms = $allowed_terms[$term->bundle()];
        foreach ($vocabulary_fields as $field) {
          if ($entity->hasField($field)) {
            $field_values = $entity->get($field)->getValue();
            if (!array_intersect(array_column($field_values, 'target_id'), $vocabulary_terms)) {
              unset($items[$item_id]);
              break;
            }
          }
        }
      }
    }
  }

  /**
   * Get field machine names associated to vocabularies.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string $bundle
   *   The entity bundle.
   * @param array $vocabularies
   *   The vocabularies to filter.
   *
   * @return array
   *   The field machine names grouped by vocabulary.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getFieldNamesFromVocabularies(string $entity_type_id, string $bundle, array $vocabularies): array {
    $fields = $this->entityTypeManager->getStorage('field_config')->loadByProperties([
      'entity_type' => $entity_type_id,
      'bundle' => $bundle,
    ]);
    $field_names = [];
    foreach ($fields as $field) {
      $field_settings = $field->getSettings();
      if ($field->getType() === 'entity_reference' && $field_settings['target_type'] === 'taxonomy_term') {
        foreach ($vocabularies as $vocabulary) {
          if (isset($field_settings['handler_settings']['target_bundles'][$vocabulary])) {
            $field_names[$vocabulary][] = $field->getName();
          }
        }
      }
    }

    return $field_names;
  }

}
